package com.cs707.AppGrader;

import android.app.Service;
import android.content.Intent;
import android.os.IBinder;
import android.os.Bundle;

import android.telephony.TelephonyManager;

import android.util.Log;

import org.apache.http.HttpResponse;
import org.apache.http.NameValuePair;
import org.apache.http.client.ClientProtocolException;
import org.apache.http.client.HttpClient;
import org.apache.http.client.entity.UrlEncodedFormEntity;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.message.BasicNameValuePair;
import org.apache.http.util.EntityUtils;


import android.widget.TextView;
import android.widget.ScrollView;

import android.app.ActivityManager;

import java.io.*;
import java.util.*;

public class AppGrader extends Service
{
    ActivityManager activityManager;
    TextView outText;

    //seconds that have gone by
    long ticks = 0;

    //how often we should take a snapshot
    long snapshotInterval =  15;

    //how often we should post data
    long postInterval =  60*60;

    //last snapshot that was taken
    Snap oldSnap = null;

    //list of stats for all applications
    AppStatList appStatList = new AppStatList();

    //grades file
    String gradesFile = "/sdcard/cs707/grades.txt";

    //if this file is created, it means we're supposed to post
    String postFile = "/sdcard/cs707/postnow";

    //true if postFile existed last time
    Boolean postFileExists = false;


    Timer timer;

    static String readFromFile(String file)
    {
	String contents = "";
	FileReader fr = null;

	try {
	    fr = new FileReader(file);
	} catch (FileNotFoundException e) {
	    return "";
	}

	Scanner sc = new Scanner(fr);        
	while(sc.hasNext()) {
	    contents += sc.nextLine() + "\n";
	}

	return contents;
    }


    public boolean writeToFile(String fileName, String data){ 
	try
	    {
		StringBuilder strbldr = new StringBuilder();
		File gpxfile = new File(fileName);
		//NOTE: versions of this function in our other 
		//applications do an append, but this just overwrites.
		FileWriter gpxwriter = new FileWriter(gpxfile,false);
		BufferedWriter out = new BufferedWriter(gpxwriter);
		out.write(strbldr.append(data).toString());
		out.close();
	    }
	catch(Exception e)
	    {
		e.printStackTrace();
		return false;
	    }

	return true;
    }



    void dumpStats() {
	/*
	outText.setText("DUMP:\n");
	for(AppStat appStat : appStatList.appStats.values()) {
	    outText.append(appStat.appName + " ");
	    outText.append(appStat.cpuSeconds + " ");
	    outText.append(appStat.memSeconds + " ");
	    outText.append(appStat.fgSeconds + " ");
	    outText.append(appStat.bgSeconds + "\n");
	}
	*/

	String text = "";

	for(AppStat appStat : appStatList.appStats.values()) {
	    text += appStat.appName + " ";
	    text += appStat.cpuSeconds + " ";
	    text += appStat.memSeconds + " ";
	    text += appStat.fgSeconds + " ";
	    text += appStat.bgSeconds + "\n";
	}

	writeToFile(gradesFile, text);
    }



    void initStats() {
	String contents = readFromFile(gradesFile);
	if(contents == "") return;
	String lines[] = contents.split("\n");

	for(String line : lines) {
	    String fields[] = line.split(" ");
	    
	    String appName = fields[0];
	    long cpuSeconds = Long.parseLong(fields[1]);
	    long memSeconds = Long.parseLong(fields[2]);
	    long fgSeconds = Long.parseLong(fields[3]);
	    long bgSeconds = Long.parseLong(fields[4]);

	    appStatList.incCPUSeconds(appName, cpuSeconds);
	    appStatList.incMemSeconds(appName, memSeconds);
	    appStatList.incFGSeconds(appName, fgSeconds);
	    appStatList.incBGSeconds(appName, bgSeconds);
	}
    }



    /** Called when the activity is first created. */
    @Override
	public void onCreate()
    {
        super.onCreate();

	//load in the previous stats from grades.txt
	initStats();

	//setup gui
	outText = new TextView(this);
	ScrollView sv = new ScrollView(this);
	sv.addView(outText);
        //setContentView(sv);


	//initialize activity manager
	activityManager = (ActivityManager)this.getSystemService(ACTIVITY_SERVICE);


	//start timer
	timer = new Timer();
	timer.scheduleAtFixedRate(
				  new TimerTask() {
				      public void run() {
					  try {
					      tick();
					  } catch (Exception ex) {
					      StackTraceElement st[] = ex.getStackTrace();
					      for(int i = 0; i < st.length; i++) {
						  outText.append(st[i].toString()+"\n");
					      }
					  }
				      }
				  },
				  0,
				  1000);
    }


    @Override
    public IBinder onBind(Intent intent) {
	// TODO Auto-generated method stub
	return null;
    }


    public void tick() {
	//get list of all processes
	List<ActivityManager.RunningAppProcessInfo> processes;
	processes = activityManager.getRunningAppProcesses();

	//loop over all processes
	for (ActivityManager.RunningAppProcessInfo procInfo : processes) {
	    String appName = procInfo.processName;

	    if(procInfo.importance == ActivityManager.RunningAppProcessInfo.IMPORTANCE_FOREGROUND) {
		//the app is in the foreground
		appStatList.incFGSeconds(appName, 1);
	    } else {
		//the app is in the background
		appStatList.incBGSeconds(appName, 1);
	    }
	}


	if((ticks % snapshotInterval) == 0) {
	    Snap newSnap = Snap.take(processes);

	    if(oldSnap != null) {
		//loop over all processes
		for (ActivityManager.RunningAppProcessInfo procInfo : processes) {
		    String appName = procInfo.processName;
		    
		    ProcessSnap newps = newSnap.ps.get(appName);
		    ProcessSnap oldps = oldSnap.ps.get(appName);

		    if(newps != null && oldps != null) {
			long cpuSecs = 
			    100 * snapshotInterval *
			    (newps.tottime - oldps.tottime) /
			    (newSnap.ss.tottime - oldSnap.ss.tottime);

			outText.append("Adding " + newps.tottime + " - " + oldps.tottime + " / " + (newSnap.ss.tottime - oldSnap.ss.tottime) + "\n");

			appStatList.incCPUSeconds(appName, cpuSecs);

			appStatList.incMemSeconds(appName, newps.rss);
		    }
		}

		dumpStats();
	    }

	    oldSnap = newSnap;
	}

	
	File f = new File(postFile);

	if((ticks % postInterval) == 0 || (f.exists() && !postFileExists)) {
	    sendPost();
	}

	postFileExists = f.exists();
	f.delete();


	//increment second count
	ticks++;
    }




    private String sendPost()
    {
	TelephonyManager tm = (TelephonyManager)getSystemService(TELEPHONY_SERVICE);
    	String responseString = "";
    	HttpClient httpclient = new DefaultHttpClient();
	HttpPost httppost = new HttpPost("http://pages.cs.wisc.edu/~harter/android/insert.php");

	String tableName = "app_grades";

	String data = "appName,cpuSeconds,memSeconds,fgSeconds,bgSeconds@@@";

	for(AppStat appStat : appStatList.appStats.values()) {
	    data += "'" + appStat.appName + "'" + ",";
	    data += appStat.cpuSeconds + ",";
	    data += appStat.memSeconds + ",";
	    data += appStat.fgSeconds + ",";
	    data += appStat.bgSeconds + "@@@";
	}



	try 
	    {
		List<NameValuePair> nameValuePairs = new ArrayList<NameValuePair>();

		nameValuePairs.add(new BasicNameValuePair("table", tableName));
		nameValuePairs.add(new BasicNameValuePair("imei", tm.getDeviceId()));
		nameValuePairs.add(new BasicNameValuePair("data", data));

		httppost.setEntity(new UrlEncodedFormEntity(nameValuePairs));
		    
		HttpResponse response = httpclient.execute(httppost);

		responseString = EntityUtils.toString(response.getEntity());

		//Log.v("cs707", responseString);
	    } 
	catch (ClientProtocolException e) 
	    {
	    } 
	catch (IOException e) 
	    {
	    }
		
	return responseString;
    }



    static class AppStatList {
	//all application stats
	Hashtable<String, AppStat> appStats = new Hashtable<String, AppStat>();
	
	void init(String appName) {
	    if(appStats.get(appName) == null) {
		AppStat as = new AppStat();
		as.appName = appName;

		appStats.put(appName, as);
	    }
	}

	void incCPUSeconds(String appName, long val) {
	    init(appName);
	    AppStat as = appStats.get(appName);
	    as.cpuSeconds += val;
	}

	void incMemSeconds(String appName, long val) {
	    init(appName);
	    AppStat as = appStats.get(appName);
	    as.memSeconds += val;
	}

	void incFGSeconds(String appName, long val) {
	    init(appName);
	    AppStat as = appStats.get(appName);
	    as.fgSeconds += val;
	}

	void incBGSeconds(String appName, long val) {
	    init(appName);
	    AppStat as = appStats.get(appName);
	    as.bgSeconds += val;
	}
    }



    static class AppStat {
	String appName;
	long cpuSeconds = 0;
	long memSeconds = 0;
	long fgSeconds = 0;
	long bgSeconds = 0;
    }


    
    static class Snap {
	//snapshots of all processes
	//indexed by proc name
	Hashtable<String, ProcessSnap> ps = new Hashtable<String, ProcessSnap>();

	SystemSnap ss;


	public static Snap take(List<ActivityManager.RunningAppProcessInfo> processes) {
	    Snap s = new Snap();

	    //take system snapshot
	    s.ss = SystemSnap.take();
	    if(s.ss == null) {
		return null;
	    }

	    //loop over all processes
	    for (ActivityManager.RunningAppProcessInfo procInfo : processes) {
		//take process snapshot
		ProcessSnap tmp = ProcessSnap.take(procInfo.pid);

		if(tmp != null) {
		    s.ps.put(procInfo.processName, tmp);
		}
	    }

	    return s;
	}
    }

    

    static class ProcessSnap {
	long utime;
	long stime;

	//sum of above fields
	long tottime;

	long vss;
	long rss;

	public void calcTotTime() {
	    this.tottime = this.utime + this.stime;
	}

	public static ProcessSnap take(int pid) {
	    String contents = readFromFile("/proc/" + pid + "/stat");
	    StringTokenizer st = new StringTokenizer(contents);
	    ProcessSnap ps = new ProcessSnap();

	    if (st.countTokens() != 44) return null;

	    st.nextToken(); //pid
	    st.nextToken(); //comm
	    st.nextToken(); //state
	    st.nextToken(); //ppid
	    st.nextToken(); //pgrp
	    st.nextToken(); //session
	    st.nextToken(); //tty_nr
	    st.nextToken(); //tpgid
	    st.nextToken(); //flags
	    st.nextToken(); //minflt
	    st.nextToken(); //cminflt
	    st.nextToken(); //majflt
	    st.nextToken(); //cmajflt
	    ps.utime = Long.parseLong(st.nextToken()); //utime
	    ps.stime = Long.parseLong(st.nextToken()); //stime
	    st.nextToken(); //cutime
	    st.nextToken(); //cstime
	    st.nextToken(); //priority
	    st.nextToken(); //nice
	    st.nextToken(); //num_threads
	    st.nextToken(); //itrealvalue
	    st.nextToken(); //starttime
	    ps.vss = Long.parseLong(st.nextToken()); //vsize
	    ps.rss = Long.parseLong(st.nextToken()); //rss
	    st.nextToken(); //rsslim
	    st.nextToken(); //startcode
	    st.nextToken(); //endcode
	    st.nextToken(); //startstack
	    st.nextToken(); //kstkesp
	    st.nextToken(); //kstkeip
	    st.nextToken(); //signal
	    st.nextToken(); //blocked
	    st.nextToken(); //sigignore
	    st.nextToken(); //sigcatch
	    st.nextToken(); //wchan
	    st.nextToken(); //nswap
	    st.nextToken(); //cnswap
	    st.nextToken(); //exit_signal
	    st.nextToken(); //processor
	    st.nextToken(); //rt_priority
	    st.nextToken(); //policy
	    st.nextToken(); //delayacct_blkio_ticks
	    st.nextToken(); //guest_time
	    st.nextToken(); //cguest_time

	    ps.calcTotTime();

	    return ps;
	}
    }

    static class SystemSnap {
	long utime;
	long ntime;
	long stime;
	long itime;
	long iotime;
	long irqtime;
	long sirqtime;

	//sum of above fields
	long tottime;

	public void calcTotTime() {
	    tottime = utime + ntime + stime + itime + iotime + irqtime + sirqtime;
	}


	public static SystemSnap take() {
	    String contents = readFromFile("/proc/stat");
	    StringTokenizer st = new StringTokenizer(contents);
	    SystemSnap ss = new SystemSnap();

	    if (st.countTokens() < 8) return null;

	    st.nextToken(); //cpu
	    ss.utime = Long.parseLong(st.nextToken()); //utime 
	    ss.ntime = Long.parseLong(st.nextToken()); //ntime 
	    ss.stime = Long.parseLong(st.nextToken()); //stime
	    ss.itime = Long.parseLong(st.nextToken()); //itime
	    ss.iotime = Long.parseLong(st.nextToken()); //iowtime 
	    ss.irqtime = Long.parseLong(st.nextToken()); //irqtime 
	    ss.sirqtime = Long.parseLong(st.nextToken()); //sirqtime

	    ss.calcTotTime();

	    return ss;
	}
    }
}
